/*
**  Copyright (c) 2007-2009 Sendmail, Inc. and its suppliers.
**	All rights reserved.
**  Sendmail, Inc. Confidential
**
**  $Id: batv-filter.c,v 1.67 2009/05/27 16:16:45 msk Exp $
*/

#ifndef lint
static char batv_filter_c_id[] = "@(#)$Id: batv-filter.c,v 1.67 2009/05/27 16:16:45 msk Exp $";
#endif /* !lint */

/* system includes */
#include <sys/param.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <sys/stat.h>
#ifdef __linux__
# include <sys/prctl.h>
#endif /* __linux__ */
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <syslog.h>
#include <errno.h>
#include <sysexits.h>
#include <fcntl.h>
#include <stdio.h>
#include <signal.h>
#include <assert.h>
#include <regex.h>
#include <ctype.h>
#include <time.h>
#include <pwd.h>
#ifdef DEBUG
# include <netdb.h>
#endif /* DEBUG */
#ifndef SOLARIS
# include <paths.h>
#else /* SOLARIS */
# if (SOLARIS >= 21000)
#  include <fenv.h>
# else /* (SOLARIS >= 21000) */
#  include <floatingpoint.h>
# endif /* (SOLARIS >= 21000) */
# include <iso/limits_iso.h>
#endif /* ! SOLARIS */

/* openssl includes */
#include <openssl/sha.h>

/* libsm includes */
#include <sm/string.h>

/* libmilter includes */
#ifndef DEBUG
#include "libmilter/mfapi.h"
#endif /* !DEBUG */

/* batv-filter includes */
#include "batv-filter.h"
#include "config.h"
#include "batv-config.h"
#include "util.h"

/* macros */
#ifndef MIN
# define MIN(x,y)	((x) < (y) ? (x) : (y))
#endif /* ! MIN */

#ifdef SOLARIS
# ifndef INADDR_NONE
#  define INADDR_NONE	(in_addr_t)(-1)
# endif /* INADDR_NONE */
#endif /* SOLARIS */

#define BATV_MODE_SIGN		0x01
#define	BATV_MODE_VERIFY	0x02
#define BATV_MODE_DEFAULT	(BATV_MODE_SIGN | BATV_MODE_VERIFY)

#ifdef DEBUG
/* DEBUGGING STUFF */
# define MI_SUCCESS	1
# define MI_FAILURE	(-1)
# define SMFIS_CONTINUE	0
# define SMFIS_ACCEPT	1
# define SMFIS_REJECT	2
# define SMFIS_DISCARD	3
# define SMFIS_TEMPFAIL	4
# define SMFIF_NONE	0x00000000L
# define SMFIF_ADDHDRS	0x00000001L
# define SMFIF_CHGBODY	0x00000002L
# define SMFIF_MODBODY	SMFIF_CHGBODY
# define SMFIF_ADDRCPT	0x00000004L
# define SMFIF_DELRCPT	0x00000008L
# define SMFIF_CHGHDRS	0x00000010L
# define SMFIF_QUARANTINE 0x00000020L
# define SMFIF_CHGFROM	0x00000040L
# define SMFIF_ADDRCPT_PAR	0x00000080L
# define SMFIF_SETSYMLIST	0x00000100L
# define SMFIP_NOCONNECT 0x00000001L
# define SMFIP_NOHELO	0x00000002L
# define SMFIP_NOMAIL	0x00000004L
# define SMFIP_NORCPT	0x00000008L
# define SMFIP_NOBODY	0x00000010L
# define SMFIP_NOHDRS	0x00000020L
# define SMFIP_NOEOH	0x00000040L
# define SMFIP_NR_HDR	0x00000080L
# define SMFIP_NOHREPL	SMFIP_NR_HDR
# define SMFIP_NOUNKNOWN 0x00000100L
# define SMFIP_NODATA    0x00000200L
# define SMFIP_SKIP	0x00000400L
# define SMFIP_RCPT_REJ	0x00000800L
# define SMFIP_NR_CONN	0x00001000L
# define SMFIP_NR_HELO	0x00002000L
# define SMFIP_NR_MAIL	0x00004000L
# define SMFIP_NR_RCPT	0x00008000L
# define SMFIP_NR_DATA	0x00010000L
# define SMFIP_NR_UNKN	0x00020000L
# define SMFIP_NR_EOH	0x00040000L
# define SMFIP_NR_BODY	0x00080000L
# define SMFIP_HDR_LEADSPC 0x00100000L
# define sfsistat	int
# define SMFICTX	void
# define _SOCK_ADDR	struct sockaddr

int smfi_insheader __P((void *, int, char *, char *));
int smfi_chgfrom __P((void *, char *, char *));
void *smfi_getpriv __P((void *));
char *smfi_getsymval __P((void *, char *));
int smfi_opensocket __P((int));
int smfi_quarantine __P((void *, char *));
void smfi_setconn __P((char *));
void smfi_setdbg __P((int));
void smfi_setpriv __P((void *, void *));
int smfi_setreply __P((void *, char *, char *, char *));

char *smfis_ret[] = {
	"SMFIS_CONTINUE",
	"SMFIS_ACCEPT",
	"SMFIS_REJECT",
	"SMFIS_DISCARD",
	"SMFIS_TEMPFAIL",
};

static void *fakepriv;				/* fake private space */
#endif /* DEBUG */

/*
**  RCPTLIST -- recipient list (old and new)
*/

struct batv_rcptlist
{
	bool		rcpt_signed;		/* signature present */
	bool		rcpt_verified;		/* signature verified */
	bool		rcpt_badversion;	/* version mismatch */
	char *		rcpt_orig;		/* original recipient */
	char *		rcpt_new;		/* replacement */
	struct batv_rcptlist * rcpt_next;	/* next node */
};
typedef struct batv_rcptlist * RCPTLIST;

/*
**  CONTEXT -- thread context
*/

struct batv_context
{
	bool		ctx_authed;		/* authenticated client */
	char *		ctx_jobid;		/* job ID */
	char *		ctx_sender;		/* envelope sender */
	char *		ctx_newsender;		/* new envelope sender */
	char *		ctx_esmtp;		/* ESMTP stuff */
	RCPTLIST	ctx_rcpts;		/* recipients */
};
typedef struct batv_context * CONTEXT;

/*
**  RELIST -- regular expression list
*/

struct batv_relist
{
	regex_t		re_re;			/* regex_t */
	struct batv_relist * re_next;		/* next node */
};
typedef struct batv_relist * RELIST;

/*
**  HOSTLIST -- hostname/CIDR list
*/

struct batv_hostlist
{
	char *		host_name;		/* regex_t */
	struct in_addr	host_addr;		/* address */
	struct in_addr	host_mask;		/* mask */
	struct batv_hostlist * host_next;	/* next node */
};
typedef struct batv_hostlist * HOSTLIST;

/*
**  DOMLIST -- domains to sign/verify
*/

struct batv_domlist
{
	char *		dom_name;		/* name */
	struct batv_domlist * dom_next;		/* next node */
};
typedef struct batv_domlist * DOMLIST;

/*
**  CONNCTX -- connection context, containing thread-specific data
*/

struct connctx
{
	bool		cctx_milterv2;		/* milter v2 available */
	bool		cctx_noleadspc;		/* no leading spaces */
	bool		cctx_internal;		/* internal host */
	char		cctx_mtaname[MAXHOSTNAMELEN + 1];
						/* MTA name */
	char		cctx_hostname[MAXHOSTNAMELEN + 1];
						/* client hostname */
	_SOCK_ADDR	cctx_addr;		/* client address */
	CONTEXT		cctx_msgctx;		/* message context */
};
typedef struct connctx * connctx;

/*
**  LOOKUP -- lookup table
*/

struct lookup
{
	char *		lu_str;
	int		lu_code;
};

/* globals */
bool dolog;
bool logwhy;
bool die;
bool addxhdr;
bool addarhdr;
bool notreally;
bool bounceonly;
bool skipauth;
bool setreply;
bool sendmail;
bool quarantine;
bool authservjobid;
int diesig;
unsigned int opmode;
unsigned int version;
unsigned int siglife;
Peer peerlist;
char *progname;
char *key;
char *authservid;
char **macros;
char **mtas;
char **values;
char hostname[MAXHOSTNAMELEN + 1];
RELIST allow;
HOSTLIST signlist;
DOMLIST domains;

struct lookup log_facilities[] =
{
	{ "auth",		LOG_AUTH },
	{ "cron",		LOG_CRON },
	{ "daemon",		LOG_DAEMON },
	{ "kern",		LOG_KERN },
	{ "lpr",		LOG_LPR },
	{ "mail",		LOG_MAIL },
	{ "news",		LOG_NEWS },
	{ "security",		LOG_AUTH },       /* DEPRECATED */
	{ "syslog",		LOG_SYSLOG },
	{ "user",		LOG_USER },
	{ "uucp",		LOG_UUCP },
	{ "local0",		LOG_LOCAL0 },
	{ "local1",		LOG_LOCAL1 },
	{ "local2",		LOG_LOCAL2 },
	{ "local3",		LOG_LOCAL3 },
	{ "local4",		LOG_LOCAL4 },
	{ "local5",		LOG_LOCAL5 },
	{ "local6",		LOG_LOCAL6 },
	{ "local7",		LOG_LOCAL7 },
	{ NULL,			-1 }
};

/*
**  ============================= LOCAL STUFF =============================
*/

/*
**  BATVF_RESTART_CHECK -- initialize/check restart rate information
**
**  Parameters:
**  	n -- size of restart rate array to initialize/enforce
**  	t -- maximum time range for restarts (0 == init)
**
**  Return value:
**  	true -- OK to continue
**  	false -- error
*/

static bool
batvf_restart_check(int n, time_t t)
{
	static int idx;				/* last filled slot */
	static int alen;			/* allocated length */
	static time_t *list;

	if (t == 0)
	{
		alen = n * sizeof(time_t);

		list = (time_t *) malloc(alen);

		if (list == NULL)
			return FALSE;

		memset(list, '\0', alen);

		idx = 0;
		alen = n;

		return TRUE;
	}
	else
	{
		int which;

		time_t now;

		(void) time(&now);

		which = (idx - 1) % alen;
		if (which == -1)
			which = alen - 1;

		if (list[which] != 0 &&
		    list[which] + t > now)
			return false;

		list[which] = t;
		idx++;

		return TRUE;
	}
}

/*
**  BATV_INIT_SYSLOG -- initialize syslog()
**
**  Parameters:
**  	facility -- name of the syslog facility to use when logging;
**  	            can be NULL to request the default
**
**  Return value:
**  	None.
*/

static void
batv_init_syslog(char *facility)
{
#ifdef LOG_MAIL
	int code;
	struct lookup *p = NULL;

	closelog();

	code = LOG_MAIL;
	if (facility != NULL)
	{
		for (p = log_facilities; p != NULL; p++)
		{
			if (strcasecmp(p->lu_str, facility) == 0)
			{
				code = p->lu_code;
				break;
			}
		}
	}

	openlog(progname, LOG_PID, code);
#else /* LOG_MAIL */
	closelog();

	openlog(progname, LOG_PID);
#endif /* LOG_MAIL */
}

/*
**  BATV_SIGHANDLER -- signal handler
**
**  Parameters:
**  	sig -- signal received
**
**  Return value:
**  	None.
*/

static void
batv_sighandler(int sig)
{
	if (sig == SIGINT || sig == SIGTERM || sig == SIGHUP)
	{
		diesig = sig;
		die = TRUE;
	}
}

/*
**  BATV_STDIO -- set up the base descriptors to go nowhere
**
**  Parameters:
**  	None.
**
**  Return value:
**  	None.
*/

static void
batv_stdio(void)
{
	int devnull;

	/* this only fails silently, but that's OK */
	devnull = open(_PATH_DEVNULL, O_RDWR, 0);
	if (devnull != -1)
	{
		(void) dup2(devnull, 0);
		(void) dup2(devnull, 1);
		(void) dup2(devnull, 2);
		if (devnull > 2)
			(void) close(devnull);
	}

	(void) setsid();
}

/*
**  BATV_KILLCHILD -- kill child process
**
**  Parameters:
**  	pid -- process ID to signal
**  	sig -- signal to use
**  	dolog -- log it?
**
**  Return value:
**  	None.
*/

static void
batv_killchild(pid_t pid, int sig)
{
	if (kill(pid, sig) == -1 && dolog)
	{
		syslog(LOG_ERR, "kill(%d, %d): %s", pid, sig,
		       strerror(errno));
	}
}

/*
**  BATV_INITCONTEXT -- initialize a context
**
**  Parameters:
**  	None.
**
**  Return value:
**  	A new CONTEXT handle, or NULL on error.
*/

static CONTEXT
batv_initcontext(void)
{
	CONTEXT new;

	new = (CONTEXT) malloc(sizeof *new);
	if (new == NULL)
		return NULL;

	memset(new, '\0', sizeof *new);

	return new;
}

#define	CLOBBER(x)	if ((x) != NULL) { \
				free((x)); \
				(x) = NULL; \
			}

/*
**  BATV_CLEANUP -- clean up a filter context
**
**  Parameters:
**  	cfc -- CONTEXT handle to be cleaned up
**  	msg -- message stuff only, or the whole shebang?
**
**  Return value:
**  	None.
*/

static void
batv_cleanup(connctx cc, bool msg)
{
	RCPTLIST rcpt;
	RCPTLIST next;
	CONTEXT bfc;

	assert(cc != NULL);

	bfc = cc->cctx_msgctx;
	if (bfc == NULL)
		return;

	CLOBBER(bfc->ctx_jobid);
	CLOBBER(bfc->ctx_sender);
	CLOBBER(bfc->ctx_newsender);
	CLOBBER(bfc->ctx_esmtp);

	rcpt = bfc->ctx_rcpts;
	while (rcpt != NULL)
	{
		next = rcpt->rcpt_next;

		CLOBBER(rcpt->rcpt_orig);
		CLOBBER(rcpt->rcpt_new);

		free(rcpt);

		rcpt = next;
	}

	free(bfc);

	cc->cctx_msgctx = NULL;
}

/*
**  BATV_STRIPBRACKETS -- remove angle brackets from the sender address
**
**  Parameters:
** 	addr -- address to be processed
**
**  Return value:
**  	None.
*/

static void
batv_stripbrackets(char *addr)
{
	char *p, *q;

	assert(addr != NULL);

	p = addr;
	q = addr + strlen(addr) - 1;

	while (*p == '<' && *q == '>')
	{
		p++;
		*q-- = '\0';
	}

	if (p != addr)
	{
		for (q = addr; *p != '\0'; p++, q++)
			*q = *p;
		*q = '\0';
	}
}

/*
**  BATV_MKREGEXP -- make a regular expression
**
**  Parameters:
**  	str -- input string
**  	re -- destination string
**
**  Return value:
**   	None.
*/

static void
batv_mkregexp(str, addr, len)
	char *str;
	char *addr;
	size_t len;
{
	char *p;
	char *q;
	char *end;

	assert(str != NULL);
	assert(addr != NULL);

	memset(addr, '\0', len);

	addr[0] = '^';

	end = addr + len - 1;

	for (p = str, q = addr + 1; *p != '\0' && q <= end; p++)
	{
		switch (*p)
		{
		  case '*':
			if (q >= end - 2)
				return;
			*q++ = '.';
			*q++ = '*';
			break;

		  case '.':
			if (q >= end - 2)
				return;
			*q++ = '\\';
			*q++ = '.';
			break;

		  default:
			*q++ = *p;
			break;
		}
	}

	if (q < end)
		*q = '$';
}

/*
**  BATV_ISBLANK -- blank line?
**
**  Parameters:
**  	str -- string
**
**  Return value:
**  	TRUE iff "str" is either zero-length or contains only spaces.
*/

static bool
batv_isblank(char *str)
{
	char *p;

	assert(str != NULL);

	for (p = str; *p != '\0'; p++)
	{
		if (!isascii(*p) || !isspace(*p))
			return FALSE;
	}

	return TRUE;
}

/*
**  ============================= DEBUG STUFF =============================
*/

#ifdef DEBUG
int
smfi_chgfrom(void *ctx, char *from, char *esmtp)
{
	printf("smfi_chgfrom(<ctx>, `%s', `%s')\n", from,
	       esmtp == NULL ? "(null)" : esmtp);
	return MI_SUCCESS;
}

int
smfi_insheader(void *ctx, int idx, char *hdr, char *val)
{
	printf("smfi_insheader(<ctx>, %d, `%s', `%s')\n", idx, hdr, val);
	return MI_SUCCESS;
}

int
smfi_addrcpt(void *ctx, char *rcpt)
{
	printf("smfi_addrcpt(<ctx>, `%s')\n", rcpt);
	return MI_SUCCESS;
}

int
smfi_delrcpt(void *ctx, char *rcpt)
{
	printf("smfi_delrcpt(<ctx>, `%s')\n", rcpt);
	return MI_SUCCESS;
}

int
smfi_opensocket(int param)
{
	printf("smfi_opensocket(%d)\n", param);

	return MI_SUCCESS;
}

int
smfi_quarantine(void *ctx, char *reason)
{
	printf("smfi_quarantine(<ctx>, `%s')\n", reason);

	return MI_SUCCESS;
}

void
smfi_setdbg(int param)
{
	printf("smfi_setdbg(%d)\n", param);
}

void
smfi_setconn(char *file)
{
	printf("smfi_setconn(`%s')\n", file);
}

void
smfi_setpriv(void *ctx, void *priv)
{
	fakepriv = priv;
}

void *
smfi_getpriv(void *ctx)
{
	return fakepriv;
}

char *
smfi_getsymval(void *ctx, char *sym)
{
	char *ret;
	size_t l;
	CONTEXT cc;

	l = strlen(sym) + 6 + 1;
	cc = fakepriv;

	printf("smfi_getsymval(<ctx>, `%s')\n", sym);
	ret = malloc(l);
	snprintf(ret, l, "DEBUG-%s", sym);
	return ret;
}

int
smfi_setreply(void *ctx, char *smtp, char *esc, char *txt)
{
	printf("smfi_setreply(<ctx>, `%s', `%s', `%s')\n", smtp, esc, txt);

	return MI_SUCCESS;
}
#endif /* DEBUG */

/*
**  ============================= MILTER STUFF =============================
*/

/*
**  MLFI_NEGOTIATE -- handler called on new SMTP connection to negotiate
**                    MTA options
**
**  Parameters:
**  	ctx -- milter context
**	f0  -- actions offered by the MTA
**	f1  -- protocol steps offered by the MTA
**	f2  -- reserved for future extensions
**	f3  -- reserved for future extensions
**	pf0 -- actions requested by the milter
**	pf1 -- protocol steps requested by the milter
**	pf2 -- reserved for future extensions
**	pf3 -- reserved for future extensions
**
**  Return value:
**  	An SMFIS_* constant.
*/

static sfsistat
mlfi_negotiate(SMFICTX *ctx,
	unsigned long f0, unsigned long f1,
	SM_UNUSED(unsigned long f2),
	SM_UNUSED(unsigned long f3),
	unsigned long *pf0, unsigned long *pf1,
	unsigned long *pf2, unsigned long *pf3)
{
	unsigned long reqactions = (SMFIF_CHGFROM |
	                            SMFIF_ADDHDRS |
	                            SMFIF_DELRCPT |
	                            SMFIF_ADDRCPT );
	unsigned long protosteps = (SMFIP_NOHELO |
	                            SMFIP_NOHDRS |
	                            SMFIP_NOEOH |
	                            SMFIP_NOUNKNOWN |
	                            SMFIP_NODATA |
	                            SMFIP_NOBODY );
	connctx cc;

	/* initialize connection context */
	cc = malloc(sizeof(struct connctx));
	if (cc != NULL)
	{
		memset(cc, '\0', sizeof(struct connctx));

		smfi_setpriv(ctx, cc);
	}

	/* verify the actions we need are available */
	if (quarantine)
		reqactions |= SMFIF_QUARANTINE;
	if ((f0 & reqactions) != reqactions)
	{
		if (dolog)
		{
			syslog(LOG_ERR,
			       "mlfi_negotiate(): required milter action(s) not available (got 0x%lx, need 0x%lx)",
			       f0, reqactions);
		}

		return SMFIS_REJECT;
	}

	/* set the actions we want */
	*pf0 = reqactions;

	/* disable as many protocol steps we don't need as are available */
	*pf1 = (protosteps & f1);

# ifdef SMFIP_HDR_LEADSPC
	/* request preservation of leading spaces if possible */
	if ((f1 & SMFIP_HDR_LEADSPC) != 0)
	{
		if (cc != NULL)
		{
			cc->cctx_noleadspc = TRUE;
			*pf1 |= SMFIP_HDR_LEADSPC;
		}
	}
# endif /* SMFIP_HDR_LEADSPC */

	*pf2 = 0;
	*pf3 = 0;

	return SMFIS_CONTINUE;
}

/*
**  MLFI_CONNECT -- handle a connection announcement
**
**  Parameters:
**  	ctx -- milter context
**  	host -- hostname of the client
**  	addr -- address information for the client
**
**  Return value:
**  	An SMFIS_* constant.
*/

sfsistat
mlfi_connect(SMFICTX *ctx, char *host, _SOCK_ADDR *addr)
{
	HOSTLIST hc;
	connctx cc;
	char *dot;

	cc = smfi_getpriv(ctx);
	assert(cc != NULL);

	/* if the client is on an ignored host, then ignore it */
	if (peerlist != NULL)
	{
		/* try hostname, if available */
		if (host != NULL && host[0] != '\0' && host[0] != '[')
		{
			batvf_lowercase(host);
			if (batvf_checkhost(peerlist, host))
				return SMFIS_ACCEPT;
		}

		/* try IP address, if available */
		if (addr != NULL && addr->sa_family == AF_INET)
		{
			if (batvf_checkip(peerlist, addr))
				return SMFIS_ACCEPT;
		}
	}

	if (authservid != NULL)
	{
		sm_strlcpy(cc->cctx_mtaname, authservid,
		           sizeof cc->cctx_mtaname);
	}
	else
	{
		char *mtaname;

		mtaname = smfi_getsymval(ctx, "j");
		if (mtaname != NULL)
		{
			sm_strlcpy(cc->cctx_mtaname, mtaname,
			           sizeof cc->cctx_mtaname);
		}
		else
		{
			sm_strlcpy(cc->cctx_mtaname, MTAUNKNOWN,
			           sizeof cc->cctx_mtaname);
		}
	}

	sm_strlcpy(cc->cctx_hostname, host, sizeof cc->cctx_hostname);
	memcpy(&cc->cctx_addr, addr, sizeof cc->cctx_addr);

	/* is the client on the "internal" (sign) list? */
	for (hc = signlist; hc != NULL; hc = hc->host_next)
	{
		if (hc->host_name != NULL)
		{
			/* full hostname */
			if (strcasecmp(hc->host_name, cc->cctx_hostname) == 0)
			{
				cc->cctx_internal = TRUE;
				break;
			}

			for (dot = strchr(cc->cctx_hostname, '.');
			     dot != NULL;
			     dot = strchr(dot + 1, '.'))
			{
				if (strcasecmp(hc->host_name, dot) == 0)
				{
					cc->cctx_internal = TRUE;
					break;
				}
			}

			if (cc->cctx_internal)
				break;
		}

		if (hc->host_addr.s_addr != 0 &&
		    hc->host_mask.s_addr != 0 &&
		    cc->cctx_addr.sa_family == AF_INET)
		{
			struct sockaddr_in *in;

			in = (struct sockaddr_in *) &cc->cctx_addr;

			if ((in->sin_addr.s_addr & hc->host_mask.s_addr) ==
			    (hc->host_addr.s_addr & hc->host_mask.s_addr))
			{
				cc->cctx_internal = TRUE;
				break;
			}
		}
	}

	return SMFIS_CONTINUE;
}

/*
**  MLFI_ENVFROM -- process envelope sender
**
**  Parameters:
**  	ctx -- milter context
**  	args -- MAIL FROM argument vector
**
**  Return value:
**  	An SMFIS_* constant.
*/
        
sfsistat
mlfi_envfrom(SMFICTX *ctx, char **args)
{
	bool sign = FALSE;
	int status;
	CONTEXT bfc = NULL;
	connctx cc = NULL;
	char *jobid;
	char *domain;
	char *dot;
	RELIST ac;
	DOMLIST dc;

	cc = smfi_getpriv(ctx);
	if (cc == NULL)
	{
		if (dolog)
		{
			syslog(LOG_ERR,
			       "envelope data received before connection information; temp-failing");
		}

		return SMFIS_TEMPFAIL;
	}
	
	batv_cleanup(cc, TRUE);

	bfc = batv_initcontext();
	if (bfc == NULL)
	{
		if (dolog)
			syslog(LOG_ERR, "malloc(): %s", strerror(errno));

		return SMFIS_TEMPFAIL;
	}

	cc->cctx_msgctx = bfc;

	jobid = smfi_getsymval(ctx, "i");
	if (jobid != NULL)
	{
		bfc->ctx_jobid = strdup(jobid);
		if (bfc->ctx_jobid == NULL)
		{
			if (dolog)
			{
				syslog(LOG_ERR, "strdup(): %s",
				       strerror(errno));
			}

			return SMFIS_TEMPFAIL;
		}
	}
	else
	{
		bfc->ctx_jobid = strdup(JOBIDUNKNOWN);
	}

	/* arrange to skip verification of authenticated senders */
	if (skipauth && smfi_getsymval(ctx, "{auth_authen}") != NULL)
		bfc->ctx_authed = TRUE;

	/* copy the envelope sender and clean it */
	bfc->ctx_sender = strdup(args[0]);
	if (bfc->ctx_sender == NULL)
	{
		if (dolog)
		{
			syslog(LOG_ERR, "strdup(): %s",
			       strerror(errno));
		}

		return SMFIS_TEMPFAIL;
	}
	batv_stripbrackets(bfc->ctx_sender);
	domain = strchr(bfc->ctx_sender, '@');

	if (domain == NULL)
	{
		if (logwhy && dolog)
		{
			syslog(LOG_INFO,
			       "%s no domain found in sender; skipping",
			       bfc->ctx_jobid);
		}

		return SMFIS_CONTINUE;
	}
	else
	{
		domain++;
	}

	/* apply allow list to envelope sender */
	for (ac = allow; ac != NULL; ac = ac->re_next)
	{
		status = regexec(&ac->re_re, bfc->ctx_sender, 0, NULL, 0);
		if (status == 0)
		{
			if (logwhy && dolog)
			{
				syslog(LOG_INFO,
				       "%s sender matches allow list; skipping",
				       bfc->ctx_jobid);
			}
	
			return SMFIS_CONTINUE;
		}
		else if (status != REG_NOMATCH)
		{
			if (dolog)
			{
				char errbuf[BUFRSZ + 1];

				memset(errbuf, '\0', sizeof errbuf);
				(void) regerror(status, &ac->re_re, errbuf,
				                BUFRSZ);
				syslog(LOG_ERR, "%s: regexec(): %s",
				       bfc->ctx_jobid, errbuf);
			}

			return SMFIS_TEMPFAIL;
		}
	}

	/* is it a domain we care about? */
	if (domains == NULL)
	{
		if (logwhy && dolog)
		{
			syslog(LOG_INFO,
			       "%s signing for all domains; message will be signed",
			       bfc->ctx_jobid);
		}

		sign = TRUE;
	}
	else
	{
		sign = FALSE;
		for (dc = domains; dc != NULL; dc = dc->dom_next)
		{
			/* full hostname */
			if (strcasecmp(dc->dom_name, domain) == 0)
			{
				sign = TRUE;
				break;
			}

			for (dot = strchr(domain, '.');
			     dot != NULL;
			     dot = strchr(dot + 1, '.'))
			{
				if (strcasecmp(dc->dom_name, dot) == 0)
				{
					sign = TRUE;
					break;
				}
			}

			if (sign)
				break;
		}

		if (sign && logwhy && dolog)
		{
			syslog(LOG_INFO,
			       "%s sender matches domain list; message will be signed",
			       bfc->ctx_jobid);
		}
	}

	/* see if it came in on an authorized MSA/MTA connection */
	if (mtas != NULL && !sign)
	{
		int n;
		char *mtaname;

		mtaname = smfi_getsymval(ctx, "{daemon_name}");

		if (mtaname != NULL)
		{
			for (n = 0; mtas[n] != NULL; n++)
			{
				if (strcasecmp(mtaname, mtas[n]) == 0)
				{
					sign = TRUE;
					break;
				}
			}
		}

		if (sign && logwhy && dolog)
		{
			syslog(LOG_INFO,
			       "%s MTA name matches MTA name list; message will be signed",
			       bfc->ctx_jobid);
		}
	}

	/* now test macros, if any */
	if (macros != NULL && !sign)
	{
		bool done = FALSE;
		int n;
		char *val;
		char *p;
		char name[BUFRSZ + 1];
		char vals[BUFRSZ + 1];

		for (n = 0; !done && macros[n] != NULL; n++)
		{
			/* retrieve the macro */
			snprintf(name, sizeof name, "{%s}", macros[n]);
			val = smfi_getsymval(ctx, name);

			/* short-circuit if the macro's not set */
			if (val == NULL)
				continue;

			/* macro set and we don't care about the value */
			if (val != NULL && values[n] == NULL)
			{
				sign = TRUE;
				break;
			}

			sm_strlcpy(vals, values[n], sizeof vals);

			for (p = strtok(vals, "|");
			     !done && p != NULL;
			     p = strtok(NULL, "|"))
			{
				if (strcasecmp(val, p) == 0)
				{
					sign = TRUE;
					done = TRUE;
				}
			}
		}

		if (sign && logwhy && dolog)
		{
			syslog(LOG_INFO,
			       "%s macro value matches macro list; message will be signed",
			       bfc->ctx_jobid);
		}
	}

	/* not to be signed, not in signing mode */
	if (!sign || (opmode & BATV_MODE_SIGN) == 0)
	{
		if (logwhy && dolog)
		{
			if (!sign)
			{
				syslog(LOG_INFO,
				       "%s signing mode not selected",
				       bfc->ctx_jobid);
			}
			else
			{
				syslog(LOG_INFO, "%s signing mode disabled",
				       bfc->ctx_jobid);
			}
		}

		return SMFIS_CONTINUE;
	}

	/* sign only for internal/authenticated clients */
	if (cc->cctx_internal || bfc->ctx_authed)
	{
		int expire;
		time_t now;
		SHA_CTX sha1;
		char buf[BUFRSZ + 1];
		unsigned char digest[SHA1_DIGEST_SIZE];
		char sender[BUFRSZ + 1];

		/* continue if already signed */
		if (!sendmail && strncmp(bfc->ctx_sender, "prvs=", 5) == 0)
			return SMFIS_CONTINUE;
		if (sendmail && strstr(bfc->ctx_sender, "+prvs=") != NULL)
			return SMFIS_CONTINUE;

		(void) time(&now);
		expire = ((now / 86400) + siglife) % 1000;
		
		snprintf(buf, sizeof buf, "%1d%03d%s%s", version, expire,
		         bfc->ctx_sender, key);

		SHA1_Init(&sha1);
		SHA1_Update(&sha1, buf, strlen(buf));
		SHA1_Final(digest, &sha1);

		if (sendmail)
		{
			char *at;
			char tmp[BUFRSZ + 1];

			sm_strlcpy(tmp, bfc->ctx_sender, sizeof tmp);

			at = strchr(tmp, '@');
			if (at != NULL)
			{
				*at = '\0';

				snprintf(sender, sizeof sender,
				         "%s+prvs=%1d%03d%02X%02X%02X@%s",
				         tmp, version, expire, digest[0],
				         digest[1], digest[2], at + 1);
			}
			else
			{
				snprintf(sender, sizeof sender,
				         "%s+prvs=%1d%03d%02X%02X%02X",
				         bfc->ctx_sender, version, expire,
				         digest[0], digest[1], digest[2]);
			}
		}
		else
		{
			snprintf(sender, sizeof sender,
			         "prvs=%1d%03d%02X%02X%02X=%s",
			         version, expire, digest[0], digest[1],
			         digest[2], bfc->ctx_sender);
		}

		bfc->ctx_newsender = strdup(sender);
		if (bfc->ctx_newsender == NULL)
		{
			if (dolog)
			{
				syslog(LOG_ERR, "%s: strdup(): %s",
				       bfc->ctx_jobid, strerror(errno));
			}

			return SMFIS_TEMPFAIL;
		}
	}
	else if (logwhy && dolog)
	{
		syslog(LOG_INFO,
		       "%s client not internal and not authenticated; signing skipped",
		       bfc->ctx_jobid);
	}

	/* if we have a new sender, we have to replace the ESMTP stuff later */
	if (bfc->ctx_newsender != NULL)
	{
		int c;
		char esmtp[BUFRSZ + 1];

		memset(esmtp, '\0', sizeof esmtp);

		for (c = 1; args[c] != NULL; c++)
		{
			if (c > 1)
				sm_strlcat(esmtp, " ", sizeof esmtp);

			sm_strlcat(esmtp, args[c], sizeof esmtp);
		}

		if (esmtp[0] != '\0')
		{
			bfc->ctx_esmtp = strdup(esmtp);
			if (bfc->ctx_esmtp == NULL)
			{
				if (dolog)
				{
					syslog(LOG_ERR, "%s: strdup(): %s",
					       bfc->ctx_jobid,
					       strerror(errno));
				}

				return SMFIS_TEMPFAIL;
			}
		}
	}
	
	return SMFIS_CONTINUE;
}

/*
**  MLFI_ENVRCPT -- envelope recipient
**
**  Parameters:
**  	ctx -- milter context
**  	args -- RCPT TO argument vector
**
**  Return value:
**  	An SMFIS_* constant.
*/

sfsistat
mlfi_envrcpt(SMFICTX *ctx, char **args)
{
	bool verify;
	int status;
	CONTEXT bfc;
	connctx cc;
	char *domain;
	char *dot;
	DOMLIST dc;
	RCPTLIST rcpt;
	RELIST ac;
	char addr[BUFRSZ + 1];

	cc = smfi_getpriv(ctx);
	assert(cc != NULL);
	bfc = cc->cctx_msgctx;
	assert(bfc != NULL);

	if (bfc == NULL)
	{
		if (dolog)
		{
			syslog(LOG_ERR,
			       "envelope data received before connection information; temp-failing");
		}

		return SMFIS_TEMPFAIL;
	}

	sm_strlcpy(addr, args[0], sizeof addr);
	batv_stripbrackets(addr);
	domain = strchr(addr, '@');

	if (domain == NULL)
	{
		if (logwhy && dolog)
		{
			syslog(LOG_INFO,
			       "%s no domain found in recipient; skipping",
			       bfc->ctx_jobid);
		}

		return SMFIS_CONTINUE;
	}
	else
	{
		domain++;
	}
	
	/* apply allow list to envelope recipient */
	for (ac = allow; ac != NULL; ac = ac->re_next)
	{
		status = regexec(&ac->re_re, addr, 0, NULL, 0);
		if (status == 0)
		{
			if (logwhy && dolog)
			{
				syslog(LOG_INFO,
				       "%s recipient matches allow list; skipping",
				       bfc->ctx_jobid);
			}
	
			return SMFIS_CONTINUE;
		}
		else if (status != REG_NOMATCH)
		{
			if (dolog)
			{
				char errbuf[BUFRSZ + 1];

				memset(errbuf, '\0', sizeof errbuf);
				(void) regerror(status, &ac->re_re, errbuf,
				                BUFRSZ);
				syslog(LOG_ERR, "%s: regexec(): %s",
				       bfc->ctx_jobid, errbuf);
			}

			return SMFIS_TEMPFAIL;
		}
	}

	/* is it a domain we care about? */
	if (domains == NULL)
	{
		if (logwhy && dolog)
		{
			syslog(LOG_INFO,
			       "%s verifying for all domains; message will be verified",
			       bfc->ctx_jobid);
		}
	
		verify = TRUE;
	}
	else
	{
		verify = FALSE;
		for (dc = domains; dc != NULL; dc = dc->dom_next)
		{
			/* full hostname */
			if (strcasecmp(dc->dom_name, domain) == 0)
			{
				verify = TRUE;
				break;
			}

			for (dot = strchr(domain, '.');
			     dot != NULL;
			     dot = strchr(dot + 1, '.'))
			{
				if (strcasecmp(dc->dom_name, dot) == 0)
				{
					verify = TRUE;
					break;
				}
			}

			if (verify)
				break;
		}

		if (verify && logwhy && dolog)
		{
			syslog(LOG_INFO,
			       "%s recipient matches domain list; message will be verified",
			       bfc->ctx_jobid);
		}
	}

	/* see if it came in on an authorized MSA/MTA connection */
	if (mtas != NULL && !verify)
	{
		int n;
		char *mtaname;

		mtaname = smfi_getsymval(ctx, "{daemon_name}");

		if (mtaname != NULL)
		{
			for (n = 0; mtas[n] != NULL; n++)
			{
				if (strcasecmp(mtaname, mtas[n]) == 0)
				{
					verify = TRUE;
					break;
				}
			}
		}

		if (verify && logwhy && dolog)
		{
			syslog(LOG_INFO,
			       "%s MTA name matches MTA name list; message will be verified",
			       bfc->ctx_jobid);
		}
	}

	/* now test macros, if any */
	if (macros != NULL && !verify)
	{
		bool done = FALSE;
		int n;
		char *val;
		char *p;
		char name[BUFRSZ + 1];
		char vals[BUFRSZ + 1];

		for (n = 0; !done && macros[n] != NULL; n++)
		{
			/* retrieve the macro */
			snprintf(name, sizeof name, "{%s}", macros[n]);
			val = smfi_getsymval(ctx, name);

			/* short-circuit if the macro's not set */
			if (val == NULL)
				continue;

			/* macro set and we don't care about the value */
			if (val != NULL && values[n] == NULL)
			{
				verify = TRUE;
				break;
			}

			sm_strlcpy(vals, values[n], sizeof vals);

			for (p = strtok(vals, "|");
			     !done && p != NULL;
			     p = strtok(NULL, "|"))
			{
				if (strcasecmp(val, p) == 0)
				{
					verify = TRUE;
					done = TRUE;
				}
			}
		}

		if (verify && logwhy && dolog)
		{
			syslog(LOG_INFO,
			       "%s macro value matches macro list; message will be verified",
			       bfc->ctx_jobid);
		}
	}

	/* not to be verified, or not in verify mode */
	if (!verify || (opmode & BATV_MODE_VERIFY) == 0)
	{
		if (logwhy && dolog)
		{
			if (!verify)
			{
				syslog(LOG_INFO,
				       "%s verifying mode not selected",
				       bfc->ctx_jobid);
			}
			else
			{
				syslog(LOG_INFO, "%s verifying mode disabled",
				       bfc->ctx_jobid);
			}
		}

		return SMFIS_CONTINUE;
	}

	/* skip mail from internal hosts and auth'd clients */
	if (cc->cctx_internal || bfc->ctx_authed)
	{
		if (logwhy && dolog)
		{
			syslog(LOG_INFO,
			       "%s client not internal and not authenticated; verifying skipped",
			       bfc->ctx_jobid);
		}

		return SMFIS_CONTINUE;
	}

	/* add the recipient */
	rcpt = (RCPTLIST) malloc(sizeof *rcpt);
	if (rcpt == NULL)
	{
		if (dolog)
			syslog(LOG_ERR, "malloc(): %s", strerror(errno));

		return SMFIS_TEMPFAIL;
	}

	rcpt->rcpt_orig = strdup(addr);
	if (rcpt->rcpt_orig == NULL)
	{
		if (dolog)
			syslog(LOG_ERR, "strdup(): %s", strerror(errno));

		free(rcpt);

		return SMFIS_TEMPFAIL;
	}

	rcpt->rcpt_next = bfc->ctx_rcpts;
	bfc->ctx_rcpts = rcpt;

	/* if it's unsigned, complain */
	if ((!sendmail && strncmp(addr, "prvs=", 5) != 0) ||
	    (sendmail && strstr(addr, "+prvs=") == NULL))
	{
		/* only complain if it's a bounce, or we're being fascist */
		if (strlen(bfc->ctx_sender) == 0 || !bounceonly)
		{
			if (dolog)
			{
				syslog(LOG_ERR,
				       "%s: BATV signature not present in `%s'",
				       bfc->ctx_jobid, addr);
			}

			if (!notreally && !addarhdr)
			{
				if (setreply &&
				    smfi_setreply(ctx, BATV_REJECTSMTP,
				                  BATV_REJECTESC,
				                  "BATV signature missing") == MI_FAILURE)
				{
					if (dolog)
					{
						syslog(LOG_ERR,
						       "%s: smfi_setreply() failed",
						       bfc->ctx_jobid);
					}
				}
			}
		}
	}
	/* if it's signed, verify */
	else
	{
		int daynum;
		int sigexpire;
		int sigversion;
		time_t now;
		char *sig = &addr[5];
		char *eq;
		SHA_CTX sha1;
		char orcpt[BUFRSZ + 1];
		char tmp[BUFRSZ + 1];
		char buf[BUFRSZ + 1];
		unsigned char digest[SHA1_DIGEST_SIZE];

		rcpt->rcpt_signed = TRUE;
		rcpt->rcpt_verified = TRUE;

		if (sendmail)
		{
			char *plus;

			plus = strstr(addr, "+prvs=");
			assert(plus != NULL);

			sig = plus + 6;
		}

		if (sendmail)
		{
			char *at;
			char *plus;

			at = strchr(sig, '@');
			assert(at != NULL);

			sm_strlcpy(orcpt, addr, sizeof orcpt);

			plus = strstr(orcpt, "+prvs=");
			assert(plus != NULL);
			*plus = '\0';

			sm_strlcat(orcpt, at, sizeof orcpt);
		}
		else
		{
			eq = strchr(sig, '=');
			if (eq == NULL)
			{
				if (dolog)
				{
					syslog(LOG_ERR,
					       "%s: BATV syntax error in `%s'",
					       bfc->ctx_jobid, addr);
				}

				rcpt->rcpt_verified = FALSE;
				return SMFIS_TEMPFAIL;
			}

			sm_strlcpy(orcpt, eq + 1, sizeof orcpt);
		}

		if (!isdigit(sig[0]) ||
		    !isdigit(sig[1]) ||
		    !isdigit(sig[2]) ||
		    !isdigit(sig[3]) ||
		    !isxdigit(sig[4]) ||
		    !isxdigit(sig[5]) ||
		    !isxdigit(sig[6]) ||
		    !isxdigit(sig[7]) ||
		    !isxdigit(sig[8]) ||
		    !isxdigit(sig[9]))
		{
			if (dolog)
			{
				syslog(LOG_ERR,
				       "%s: BATV syntax error in `%s'",
				       bfc->ctx_jobid, addr);
			}

			rcpt->rcpt_verified = FALSE;
			return SMFIS_TEMPFAIL;
		}

		sigversion = sig[0] - '0';
		if (sigversion != version)
		{
			if (dolog)
			{
				syslog(LOG_ERR,
				       "%s: BATV version mismatch in `%s'",
				       bfc->ctx_jobid, addr);
			}

			if (addarhdr)
			{
				rcpt->rcpt_badversion = TRUE;
				return SMFIS_CONTINUE;
			}

			return SMFIS_TEMPFAIL;
		}

		sigexpire = (sig[1] - '0') * 100 +
		            (sig[2] - '0') * 10 +
		            (sig[3] - '0');

		memset(buf, '\0', sizeof buf);
		buf[0] = sig[0];
		buf[1] = sig[1];
		buf[2] = sig[2];
		buf[3] = sig[3];
		sm_strlcat(buf, orcpt, sizeof buf);
		sm_strlcat(buf, key, sizeof buf);

		SHA1_Init(&sha1);
		SHA1_Update(&sha1, buf, strlen(buf));
		SHA1_Final(digest, &sha1);

		if (sendmail)
		{
			char *at;
			char tmpaddr[BUFRSZ + 1];

			sm_strlcpy(tmpaddr, orcpt, sizeof tmpaddr);

			at = strchr(tmpaddr, '@');
			assert(at != NULL);
			*at = '\0';

			snprintf(tmp, sizeof tmp,
			         "%s+prvs=%1d%03d%02X%02X%02X@%s",
			         tmpaddr, sigversion, sigexpire,
			         digest[0], digest[1], digest[2], at + 1);
		}
		else
		{
			snprintf(tmp, sizeof tmp,
			         "prvs=%1d%03d%02X%02X%02X=%s",
			         sigversion, sigexpire, digest[0], digest[1],
			         digest[2], orcpt);
		}

		if (strcmp(tmp, addr) != 0)
		{
			rcpt->rcpt_verified = FALSE;

			if (dolog)
			{
				syslog(LOG_ERR,
				       "%s: BATV validation failed in `%s'",
				       bfc->ctx_jobid, addr);
			}

			if (!notreally && !addarhdr)
			{
				if (setreply &&
				    smfi_setreply(ctx, BATV_REJECTSMTP,
				                  BATV_REJECTESC,
				                  "BATV signature validation failed") == MI_FAILURE)
				{
					if (dolog)
					{
						syslog(LOG_ERR,
						       "%s: smfi_setreply() failed",
						       bfc->ctx_jobid);
					}
				}

				return SMFIS_REJECT;
			}
		}

		(void) time(&now);
		daynum = (now / 86400) % 1000;

		/* check expiration */
		if (sigexpire < siglife && daynum >= 1000 - siglife)
			daynum -= 1000;
		if (daynum > sigexpire)
		{
			rcpt->rcpt_verified = FALSE;

			if (dolog)
			{
				syslog(LOG_ERR,
				       "%s: BATV signature expired in `%s'",
				       bfc->ctx_jobid, addr);
			}

			if (!notreally && !addarhdr)
			{
				if (setreply &&
				    smfi_setreply(ctx, BATV_REJECTSMTP,
				                  BATV_REJECTESC,
				                  "BATV signature expired") == MI_FAILURE)
				{
					if (dolog)
					{
						syslog(LOG_ERR,
						       "%s: smfi_setreply() failed",
						       bfc->ctx_jobid);
					}
				}

				return SMFIS_REJECT;
			}
		}

		/* new recipient is the original recipient */
		rcpt->rcpt_new = strdup(orcpt);
		if (rcpt->rcpt_new == NULL)
		{
			if (dolog)
			{
				syslog(LOG_ERR, "%s: strdup(): %s",
				       bfc->ctx_jobid, strerror(errno));
			}

			return SMFIS_TEMPFAIL;
		}
	}

	return SMFIS_CONTINUE;
}

/*
**  MLFI_EOM -- end of message
**
**  Parameters:
**  	ctx -- milter context
**
**  Return value:
**  	SMFIS_CONTINUE
*/

sfsistat
mlfi_eom(SMFICTX *ctx)
{
	bool doquarantine = FALSE;
	CONTEXT bfc;
	RCPTLIST rcpt;
	connctx cc;

	cc = smfi_getpriv(ctx);
	assert(cc != NULL);
	bfc = cc->cctx_msgctx;
	assert(bfc != NULL);

	if (bfc->ctx_newsender != NULL)
	{
		if (notreally)
		{
			syslog(LOG_INFO,
			       "%s suppressed sender change from `%s' to `%s'",
			       bfc->ctx_jobid, bfc->ctx_sender,
			       bfc->ctx_newsender);
		}
		else if (smfi_chgfrom(ctx, bfc->ctx_newsender,
#if SMFI_VERSION >= 0x01000001
		                      bfc->ctx_esmtp) != MI_SUCCESS)
#else /* SMFI_VERSION >= 0x01000001 */
		                      NULL) != MI_SUCCESS)
#endif /* SMFI_VERSION >= 0x01000001 */
		{
			if (dolog)
			{
				syslog(LOG_ERR,
				       "%s sender change to `%s' failed",
				       bfc->ctx_jobid, bfc->ctx_newsender);
			}

			return SMFIS_TEMPFAIL;
		}
	}

	for (rcpt = bfc->ctx_rcpts; rcpt != NULL; rcpt = rcpt->rcpt_next)
	{
		assert(rcpt->rcpt_orig != NULL);

		if (!rcpt->rcpt_signed || !rcpt->rcpt_verified)
			doquarantine = TRUE;

		if (addarhdr)
		{
			char arhdr[BATV_MAXHEADER + 1];

			memset(arhdr, '\0', sizeof arhdr);

			if (rcpt->rcpt_badversion)
			{
				snprintf(arhdr, sizeof arhdr, 
				         "%s%s%s%s; x-batv=%s smtp.rcpt=%s",
				         cc->cctx_noleadspc ? " " : "",
				         cc->cctx_mtaname,
				         authservjobid ? "/" : "",
				         authservjobid ? bfc->ctx_jobid : "",
				         "fail (version mismatch)",
				         rcpt->rcpt_orig);
			}
			else if (rcpt->rcpt_signed)
			{
				snprintf(arhdr, sizeof arhdr, 
				         "%s%s%s%s; x-batv=%s smtp.rcpt=%s",
				         cc->cctx_noleadspc ? " " : "",
				         cc->cctx_mtaname,
				         authservjobid ? "/" : "",
				         authservjobid ? bfc->ctx_jobid : "",
				         rcpt->rcpt_verified ? "pass"
				                             : "fail (verification failed)",
				         rcpt->rcpt_orig);
			}
			else if (!rcpt->rcpt_signed)
			{
				snprintf(arhdr, sizeof arhdr, 
				         "%s%s%s%s; x-batv=%s smtp.rcpt=%s",
				         cc->cctx_noleadspc ? " " : "",
				         cc->cctx_mtaname,
				         authservjobid ? "/" : "",
				         authservjobid ? bfc->ctx_jobid : "",
				         "fail (not signed)", rcpt->rcpt_orig);
			}

			if (arhdr[0] != '\0' &&
			    smfi_insheader(ctx, 1, AUTHRESULTSHDR,
			                   arhdr) != MI_SUCCESS)
			{
				if (dolog)
				{
					syslog(LOG_ERR,
					       "%s \"%s\" header add failed",
					       bfc->ctx_jobid, AUTHRESULTSHDR);
				}

				return SMFIS_TEMPFAIL;
			}
		}

		if (rcpt->rcpt_new != NULL)
		{
			if (notreally)
			{
				syslog(LOG_INFO,
				       "%s suppressed recipient change from `%s' to `%s'",
				       bfc->ctx_jobid, rcpt->rcpt_orig,
				       rcpt->rcpt_new);
				continue;
			}

			if (smfi_delrcpt(ctx, rcpt->rcpt_orig) != MI_SUCCESS)
			{
				if (dolog)
				{
					syslog(LOG_ERR,
					       "%s recipient delete of `%s' failed",
					       bfc->ctx_jobid, rcpt->rcpt_orig);
				}
	
				return SMFIS_TEMPFAIL;
			}
	
			if (smfi_addrcpt(ctx, rcpt->rcpt_new) != MI_SUCCESS)
			{
				if (dolog)
				{
					syslog(LOG_ERR,
					       "%s recipient add of `%s' failed",
					       bfc->ctx_jobid, rcpt->rcpt_new);
				}
	
				return SMFIS_TEMPFAIL;
			}
		}
	}

	if (addxhdr)
	{
		char xfhdr[BATV_MAXHEADER + 1];

		memset(xfhdr, '\0', sizeof xfhdr);

		snprintf(xfhdr, BATV_MAXHEADER, "%s v%s %s %s", BATV_PRODUCT,
		         BATV_VERSION, hostname,
		         bfc->ctx_jobid != NULL ? bfc->ctx_jobid
		                                : JOBIDUNKNOWN);

		if (smfi_insheader(ctx, 1, XHEADERNAME, xfhdr) != MI_SUCCESS)
		{
			if (dolog)
			{
				syslog(LOG_ERR, "%s \"%s\" header add failed",
				       bfc->ctx_jobid, XHEADERNAME);
			}

			return SMFIS_TEMPFAIL;
		}
	}

	if (doquarantine && quarantine)
	{
		if (smfi_quarantine(ctx, QREASON) != MI_SUCCESS)
		{
			if (dolog)
			{
				syslog(LOG_ERR, "%s smfi_quarantine",
				       bfc->ctx_jobid);
			}

			return SMFIS_TEMPFAIL;
		}
	}

	return SMFIS_CONTINUE;
}

/*
**  MLFI_CLOSE -- shut down a connection
**
**  Parameters:
**  	ctx -- milter context
**
**  Return value:
**  	SMFIS_CONTINUE
*/

sfsistat
mlfi_close(SMFICTX *ctx)
{
	connctx cc;

	cc = smfi_getpriv(ctx);

	if (cc != NULL)
	{
		batv_cleanup(cc, FALSE);

		free(cc);

		smfi_setpriv(ctx, NULL);
	}

	return SMFIS_CONTINUE;
}

/*
**  ============================= GENERAL STUFF =============================
*/

/*
**  BATV_DEBUG -- debugging code; simulates libmilter calls
**
**  Parameters:
**  	None.
**
**  Return value:
**  	None.
*/

#ifdef DEBUG
int
batv_debug(void)
{
	bool done;
	int status;
	size_t len;
	time_t now;
	char *p;
	char *env[2];
	char tmpblock[4096];
	char data[513];
	char block[4096];

	time(&now);
	srandom(now);

	memset(data, '\0', sizeof data);
	memset(tmpblock, '\0', sizeof tmpblock);

	for (;;)
	{
		if (fgets(data, 512, stdin) == NULL)
			return 1;

		for (p = data; *p != '\0'; p++)
		{
			if (*p == '\r' || *p == '\n')
			{
				*p = '\0';
				break;
			}
		}

		if (strcmp(data, ".") == 0)
			break;

		env[0] = &data[1];
		env[1] = NULL;

		if (data[0] == 'C')
		{
			struct hostent *h;
			struct sockaddr_in in;
			unsigned long l0, l1, l2, l3, l4, l5, l6, l7;

			h = gethostbyname(&data[1]);
			if (h == NULL)
			{
				printf("gethostbyname(\"%s\") failed\n",
				       &data[1]);
				return 1;
			}
			in.sin_family = AF_INET;
			in.sin_port = htons(time(NULL) % 65536);
			memcpy(&in.sin_addr.s_addr, h->h_addr,
			       sizeof in.sin_addr.s_addr);

			l0 = 0xffffffffL;
			l1 = 0xffffffffL;
			l2 = 0;
			l3 = 0;
			l4 = 0;
			l5 = 0;
			l6 = 0;
			l7 = 0;
			status = mlfi_negotiate(NULL, l0, l1, l2, l3,
			                        &l4, &l5, &l6, &l7);
			printf("mlfi_negotiate(...) returns %s\n",
			       smfis_ret[status]);

			status = mlfi_connect(NULL, &data[1],
			                      (_SOCK_ADDR *) &in);
			printf("mlfi_connect(NULL, `%s', `%s') returns %s\n",
			       &data[1], inet_ntoa(in.sin_addr),
			       smfis_ret[status]);
		}
		else if (data[0] == 'F')
		{
			status = mlfi_envfrom(NULL, env);
			printf("mlfi_envfrom(NULL, `%s') returns %s\n", env[0],
	       		       smfis_ret[status]);
		}
		else if (data[0] == 'R')
		{
			status = mlfi_envrcpt(NULL, env);
			printf("mlfi_envrcpt(NULL, `%s') returns %s\n", env[0],
	       		       smfis_ret[status]);
		}
		else
		{
			printf("?\n");
			continue;
		}
		if (status != SMFIS_CONTINUE)
			return 0;
	}

	done = FALSE;
	while (!done)
	{
		len = fread(block, 1, sizeof block, stdin);

		if (len < sizeof block)
			done = TRUE;
	}

	status = mlfi_eom(NULL);
	printf("mlfi_eom(NULL) returns %s\n", smfis_ret[status]);
	return 0;
}
#endif /* DEBUG */

/*
**  USAGE -- usage message
**
**  Parameters:
**  	None.
**
**  Return value:
**  	EX_USAGE
*/

int
usage(void)
{
	fprintf(stderr, "%s: usage: %s -p socketfile -k key [options]\n"
	                "\t-a allowaddrs         \tfile containing addresses to pass\n"
	                "\t-A                    \tenable auto-restart\n"
	                "\t-c conffile           \tconfiguration file\n"
	                "\t-d domainlist         \tdomain(s) to verify\n"
	                "\t-D debuglevel         \tset milter debug level\n"
	                "\t-f                    \tdon't fork-and-exit\n"
	                "\t-i internallist       \tinternal host list\n"
	                "\t-l                    \tlog activity to system log\n"
	                "\t-m mtalist            \tMTA name(s) for which to sign\n"
	                "\t-M macro[=value][,...]\tmacro tests for signing\n"
	                "\t-n                    \tdon't actually do any filtering\n"
	                "\t-P pidfile            \tfile to which to write pid\n"
	                "\t-q                    \tquarantine messages that fail verification\n"
	                "\t-r                    \tadd header instead of filtering\n"
	                "\t-s                    \tskip mail from authenticated clients\n"
	                "\t-S                    \trequest meaningful SMTP replies when rejecting\n"
			"\t-t ttl                \tsignature time-to-live (days)\n"
			"\t-u userid             \tchange to specified userid\n"
			"\t-v version            \tkey version\n"
	                "\t-V                    \tprint version and terminate\n"
	                "\t-x                    \texperimental sendmail mode\n",
	        progname, progname);
	return EX_USAGE;
}

#ifndef DEBUG
/*
**  smfilter -- the milter module description
*/

struct smfiDesc smfilter =
{
	"Sendmail BATV Filter",	/* filter name */
	SMFI_VERSION,		/* version code -- do not change */
	SMFIF_ADDHDRS|SMFIF_CHGFROM|SMFIF_DELRCPT|SMFIF_ADDRCPT, /* flags */
	mlfi_connect,		/* connection info filter */
	NULL,			/* SMTP HELO command filter */
	mlfi_envfrom,		/* envelope sender filter */
	mlfi_envrcpt,		/* envelope recipient filter */
	NULL,			/* header filter */
	NULL,			/* end of headers */
	NULL,			/* body block filter */
	mlfi_eom,		/* end of message */
	NULL,			/* message aborted */
	mlfi_close,		/* shutdown */
	NULL,			/* unknown commands */
	NULL,			/* DATA command */
	mlfi_negotiate		/* option negotiation */
};
#endif /* ! DEBUG */

/*
**  MAIN -- program mainline
**
**  Parameters:
**  	The usual.
**
**  Return value:
**  	Exit status.
*/

int
main(int argc, char **argv)
{
	bool autorestart = FALSE;
	bool dofork = TRUE;
	bool enablecores = FALSE;
	int c;
	int dbglevel = -1;
#ifndef DEBUG
	int devnull;
#endif /* ! DEBUG */
	int status;
	int maxrestarts = 0;
	int maxrestartrate_n = 0;
	size_t rlen;
	time_t maxrestartrate_t = 0;
	FILE *f;
	char *p;
	char *conn = NULL;
	char *user = NULL;
	char *domainlist = NULL;
	char *mtanames = NULL;
	char *ilist = NULL;
	char *pidfile = NULL;
	char *allowlist = NULL;
	char *macrolist = NULL;
	char *keyfile = NULL;
	char *conffile = NULL;
	char *basedir = NULL;
	char *facility = NULL;
	struct stat s;

	progname = (p = strrchr(argv[0], '/')) == NULL ? argv[0] : p + 1;

	addxhdr = FALSE;
	addarhdr = FALSE;
	notreally = FALSE;
	quarantine = FALSE;
	die = FALSE;
	dolog = FALSE;
	logwhy = FALSE;
	authservjobid = FALSE;
	bounceonly = TRUE;
	skipauth = FALSE;
	sendmail = FALSE;
	setreply = FALSE;
	key = NULL;
	allow = NULL;
	signlist = NULL;
	authservid = NULL;
	macros = NULL;
	values = NULL;
	mtas = NULL;
	peerlist = NULL;
	version = (unsigned int) -1;
	siglife = (unsigned int) -1;
	opmode = BATV_MODE_DEFAULT;

	memset(hostname, '\0', sizeof hostname);
	gethostname(hostname, MAXHOSTNAMELEN);

	/* Process command line options */
	while ((c = getopt(argc, argv, CMDLINEOPTS)) != -1)
	{
		switch (c)
		{
		  case 'a':
			if (allowlist != NULL)
			{
				fprintf(stderr,
				        "%s: multiple use of -%c not allowed\n",
				        progname, c);
				return EX_USAGE;
			}
			allowlist = optarg;
			break;

		  case 'A':
			autorestart = TRUE;
			break;

		  case 'b':
			bounceonly = FALSE;
			break;

		  case 'c':
			if (conffile != NULL)
			{
				fprintf(stderr,
				        "%s: multiple use of -%c not allowed\n",
				        progname, c);
				return EX_USAGE;
			}
			conffile = optarg;
			break;

		  case 'd':
			if (domainlist != NULL)
			{
				fprintf(stderr,
				        "%s: multiple use of -%c not allowed\n",
				        progname, c);
				return EX_USAGE;
			}
			domainlist = optarg;
			break;

		  case 'D':
			dbglevel = (int) strtol(optarg, &p, 10);
			if (*p != '\0')
			{
				fprintf(stderr,
				        "%s: invalid debug level `%s'\n",
				        progname, optarg);
				return EX_USAGE;
			}
			smfi_setdbg(dbglevel);
			break;

		  case 'f':
			dofork = FALSE;
			break;

		  case 'h':
			addxhdr = TRUE;
			break;

		  case 'i':
			if (ilist != NULL)
			{
				fprintf(stderr,
				        "%s: multiple use of -%c not allowed\n",
				        progname, c);
				return EX_USAGE;
			}
			ilist = optarg;
			break;

		  case 'k':
			if (keyfile != NULL)
			{
				fprintf(stderr,
				        "%s: multiple use of -%c not allowed\n",
				        progname, c);
				return EX_USAGE;
			}
			keyfile = optarg;
			break;

		  case 'l':
			dolog = TRUE;
			break;

		  case 'm':
			mtanames = optarg;
			break;

		  case 'M':
			macrolist = optarg;
			break;

		  case 'n':
			notreally = TRUE;
			break;

		  case 'p':
			if (conn != NULL)
			{
				fprintf(stderr,
				        "%s: multiple use of -%c not allowed\n",
				        progname, c);
				return EX_USAGE;
			}
			conn = optarg;
			break;

		  case 'P':
			if (pidfile != NULL)
			{
				fprintf(stderr,
				        "%s: multiple use of -%c not allowed\n",
				        progname, c);
				return EX_USAGE;
			}
			pidfile = optarg;
			break;

		  case 'q':
			quarantine = TRUE;
			break;

		  case 'r':
			addarhdr = TRUE;
			break;

		  case 's':
			skipauth = TRUE;
			break;

		  case 'S':
			setreply = TRUE;
			break;

		  case 't':
			errno = 0;
			siglife = (u_int) strtoul(optarg, &p, 10);
			if (*p != '\0' || errno == ERANGE)
			{
				fprintf(stderr,
				        "%s: invalid signature TTL `%s'\n",
				        progname, optarg);
				return EX_USAGE;
			}
			break;

		  case 'u':
			if (user != NULL)
			{
				fprintf(stderr,
				        "%s: multiple use of -%c not allowed\n",
				        progname, c);
				return EX_USAGE;
			}
			user = optarg;
			break;

		  case 'v':
			errno = 0;
			version = (u_int) strtoul(optarg, &p, 10);
			if (*p != '\0' || errno == ERANGE || version > 9)
			{
				fprintf(stderr,
				        "%s: invalid key version `%s'\n",
				        progname, optarg);
				return EX_USAGE;
			}
			break;

		  case 'V':
			fprintf(stdout, "%s: %s v%s\n", progname,
			        BATV_PRODUCT, BATV_VERSION);
			return 0;

		  case 'x':
			sendmail = TRUE;
			break;

		  default:
			return usage();
		}
	}

	if (conffile != NULL)
	{
		int line;
		struct config *cfg;
		char *missing;
		char *tmp;
		char path[MAXPATHLEN + 1];

		sm_strlcpy(path, conffile, sizeof path);

		cfg = config_load(conffile, batvf_config, &line,
		                  path, sizeof path);

		if (cfg == NULL)
		{
			fprintf(stderr,
			        "%s: %s: configuration error at line %u: %s\n",
			        progname, path, line, config_error());

			return EX_CONFIG;
		}

		missing = config_check(cfg, batvf_config);
		if (missing != NULL)
		{
			if (dolog)
			{
				fprintf(stderr,
				        "%s: %s: required parameter \"%s\" missing\n",
				        progname, conffile, missing);
			}

			config_free(cfg);
			return EX_CONFIG;
		}

		/*
		**  Process config file contents.
		*/

		/* "AddAuthenticationResults" */
		if (!addarhdr)
		{
			(void) config_get(cfg, "AddAuthenticationResults",
			                  &addarhdr, sizeof addarhdr);
		}

		/* "AllSenders" */
		if (bounceonly)
		{
			(void) config_get(cfg, "AllSenders",
			                  &bounceonly, sizeof bounceonly);
		}

		/* "AuthservID" */
		(void) config_get(cfg, "AuthservID",
		                  &authservid, sizeof authservid);

		/* "AuthservIDWithJobID" */
		(void) config_get(cfg, "AuthservIDWithJobID",
		                  &authservjobid, sizeof authservjobid);

		/* "AutoRestart" */
		if (!autorestart)
		{
			(void) config_get(cfg, "AutoRestart",
			                  &autorestart, sizeof autorestart);
		}

		/* "AutoRestartCount" and "AutoRestartRate" */
		if (autorestart)
		{
			char *rate;

			(void) config_get(cfg, "AutoRestartCount",
			                  &maxrestarts, sizeof maxrestarts);

			(void) config_get(cfg, "AutoRestartRate", &rate,
			                  sizeof rate);

			if (rate != NULL)
			{
				int n;
				time_t t;
				char *q;

				p = strchr(rate, '/');
				if (p == NULL)
				{
					fprintf(stderr,
					        "%s: AutoRestartRate invalid\n",
					        progname);

					config_free(cfg);
					return EX_CONFIG;
				}

				*p = '\0';
				n = strtol(rate, &q, 10);
				if (n < 0 || *q != '\0')
				{
					fprintf(stderr,
					        "%s: AutoRestartRate invalid\n",
					        progname);

					config_free(cfg);
					return EX_CONFIG;
				}

				t = (time_t) strtoul(p + 1, &q, 10);
				switch (*q)
				{
				  case 'd':
				  case 'D':
					t *= 86400;
					break;

				  case 'h':
				  case 'H':
					t *= 3600;
					break;

				  case 'm':
				  case 'M':
					t *= 60;
					break;

				  case '\0':
				  case 's':
				  case 'S':
					break;

				  default:
					t = 0;
					break;
				}

				if (*q != '\0' && *(q + 1) != '\0')
					t = 0;

				if (t == 0)
				{
					fprintf(stderr,
					        "%s: AutoRestartRate invalid\n",
					        progname);

					config_free(cfg);
					return EX_CONFIG;
				}

				maxrestartrate_n = n;
				maxrestartrate_t = t;
			}
		}

		/* "Background" */
		if (dofork)
		{
			(void) config_get(cfg, "Background",
			                  &dofork, sizeof dofork);
		}

		/* "BaseDirectory" */
		(void) config_get(cfg, "BaseDirectory",
		                  &basedir, sizeof basedir);

		/* "Domain" */
		if (domainlist == NULL)
		{
			(void) config_get(cfg, "Domain",
			                  &domainlist, sizeof domainlist);
		}

		/* "DontSignMailTo" */
		if (allowlist == NULL)
		{
			(void) config_get(cfg, "DontSignMailTo",
			                  &allowlist, sizeof allowlist);
		}

		/* "EnableCoredumps" */
		(void) config_get(cfg, "EnableCoredumps",
		                  &enablecores, sizeof enablecores);

		/* "IgnoreAuthenticatedClients" */
		if (!skipauth)
		{
			(void) config_get(cfg, "IgnoreAuthenticatedClients",
			                  &skipauth, sizeof skipauth);
		}

		/* "InternalHosts" */
		if (ilist == NULL)
		{
			(void) config_get(cfg, "InternalHosts",
			                  &ilist, sizeof ilist);
		}

		/* "KeyFile" */
		if (keyfile == NULL)
		{
			(void) config_get(cfg, "KeyFile",
			                  &keyfile, sizeof keyfile);
		}

		/* XXX -- "KeyList" */

		/* "KeyVersion" */
		if (version == (unsigned int) -1)
		{
			(void) config_get(cfg, "KeyVersion",
			                  &version, sizeof version);

			if (version != (unsigned int) -1 && version > 9)
			{
				fprintf(stderr, "%s: %s: invalid key version\n",
				        progname, conffile);

				config_free(cfg);
				return EX_CONFIG;
			}
		}

		/* "LogWhy" */
		(void) config_get(cfg, "LogWhy", &logwhy, sizeof logwhy);

		/* "MacroList" */
		if (macrolist == NULL)
		{
			(void) config_get(cfg, "MacroList",
			                  &macrolist, sizeof macrolist);
		}

		/* "MilterDebug" */
		if (dbglevel == -1)
		{
			(void) config_get(cfg, "MilterDebug",
			                  &dbglevel, sizeof dbglevel);
			(void) smfi_setdbg(dbglevel);
		}

		/* "Mode" */
		tmp = NULL;
		(void) config_get(cfg, "Mode", &tmp, sizeof tmp);
		if (tmp != NULL)
		{
			unsigned int newmode = 0;

			for (p = tmp; *p != '\0'; p++)
			{
				switch (*p)
				{
				  case 's':
					newmode |= BATV_MODE_SIGN;
					break;

				  case 'v':
					newmode |= BATV_MODE_VERIFY;
					break;

				  default:
					fprintf(stderr,
					        "%s: %s: invalid mode\n",
					        progname, conffile);

					config_free(cfg);
					return EX_CONFIG;
				}
			}

			opmode = newmode;
		}

		/* "MTA" */
		if (mtanames == NULL)
		{
			(void) config_get(cfg, "MTA",
			                  &mtanames, sizeof mtanames);
		}

		/* "PeerList" */
		tmp = NULL;
		(void) config_get(cfg, "PeerList", &tmp, sizeof tmp);
		if (tmp != NULL)
		{
			f = fopen(tmp, "r");
			if (f == NULL)
			{
				fprintf(stderr, "%s: %s: fopen(): %s\n",
				        progname, tmp, strerror(errno));

				config_free(cfg);
				return EX_UNAVAILABLE;
			}

			if (!batvf_load_list(f, &peerlist))
			{
				fprintf(stderr,
				        "%s: failed to load PeerList from %s\n",
				        progname, tmp);

				fclose(f);

				config_free(cfg);
				return EX_UNAVAILABLE;
			}

			fclose(f);
		}

		/* "PidFile" */
		if (pidfile == NULL)
		{
			(void) config_get(cfg, "PidFile",
			                  &pidfile, sizeof pidfile);
		}

		/* "Quarantine" */
		if (!quarantine)
		{
			(void) config_get(cfg, "Quarantine",
			                  &quarantine, sizeof quarantine);
		}

		/* XXX -- "RemoveARAll" */
		/* XXX -- "RemoveARFrom" */

		/* "SendmailFormat" */
		if (!sendmail)
		{
			(void) config_get(cfg, "SendmailFormat",
			                  &sendmail, sizeof sendmail);
		}

		/* "SetReplies" */
		if (!setreply)
		{
			(void) config_get(cfg, "SetReplies",
			                  &setreply, sizeof setreply);
		}

		/* "SignatureTTL" */
		if (siglife == (unsigned int) -1)
		{
			(void) config_get(cfg, "SignatureTTL",
			                  &siglife, sizeof siglife);
		}

		/* "Socket" */
		if (conn == NULL)
			(void) config_get(cfg, "Socket", &conn, sizeof conn);

		/* "Syslog" */
		if (!dolog)
		{
			(void) config_get(cfg, "Syslog",
			                  &dolog, sizeof dolog);
		}

		/* "SyslogFacility" */
		(void) config_get(cfg, "SyslogFacility",
		                  &facility, sizeof facility);

		/* "TestMode" */
		if (!notreally)
		{
			(void) config_get(cfg, "TestMode",
			                  &notreally, sizeof notreally);
		}

		/* "UserID" */
		if (user == NULL)
			(void) config_get(cfg, "UserID", &user, sizeof user);

		/* "UserID" */
		if (!addxhdr)
		{
			(void) config_get(cfg, "X-Header",
			                  &addxhdr, sizeof addxhdr);
		}
	}

	if (enablecores)
	{
		bool enabled = FALSE;

#ifdef __linux__
		if (prctl(PR_SET_DUMPABLE, 1) == -1)
		{
			if (dolog)
			{
				syslog(LOG_ERR, "prctl(): %s",
				       strerror(errno));
			}

			fprintf(stderr, "%s: prctl(): %s\n",
			        progname, strerror(errno));
		}
		else
		{
			enabled = TRUE;
		}
#endif /* __linux__ */

		if (!enabled)
		{
			if (dolog)
			{
				syslog(LOG_WARNING,
				       "can't enable coredumps; continuing");
			}

			fprintf(stderr,
			        "%s: can't enable coredumps; continuing\n",
			        progname);
		}
	}

	/* conn is required */
	if (conn == NULL || keyfile == NULL)
		return usage();

	/* assert default version */
	if (version == (unsigned int) -1)
		version = 0;

	/* assert default signature lifetime */
	if (siglife == (unsigned int) -1)
		siglife = DEFSIGLIFETIME;

	/* change to base directory if requested */
	if (basedir != NULL)
	{
		status = chdir(basedir);
		if (status != 0)
		{
			fprintf(stderr, "%s: %s: chdir(): %s\n",
			        progname, basedir, strerror(errno));

			return EX_UNAVAILABLE;
		}
	}

	/* load the key */
	f = fopen(keyfile, "r");
	if (f == NULL)
	{
		fprintf(stderr, "%s: %s: fopen(): %s\n", progname, keyfile,
		        strerror(errno));
		return EX_UNAVAILABLE;
	}
	else if (fstat(fileno(f), &s) == -1)
	{
		fprintf(stderr, "%s: %s: fstat(): %s\n", progname, keyfile,
		        strerror(errno));
		fclose(f);
		return EX_UNAVAILABLE;
	}
	else if (s.st_size == 0)
	{
		fprintf(stderr, "%s: %s: empty file\n", progname, keyfile);
		fclose(f);
		return EX_UNAVAILABLE;
	}

	key = (char *) malloc(s.st_size);
	if (key == NULL)
	{
		fprintf(stderr, "%s: malloc(): %s\n", progname,
		        strerror(errno));
		fclose(f);
		return EX_UNAVAILABLE;
	}
	rlen = fread(key, 1, s.st_size, f);
	if (rlen < s.st_size)
	{
		if (ferror(f))
		{
			fprintf(stderr, "%s: %s: fread(): %s\n", progname,
			        keyfile, strerror(errno));
		}
		else if (feof(f))
		{
			fprintf(stderr,
			        "%s: %s: fread(): unexpected end-of-file\n",
			        progname, keyfile);
		}

		fclose(f);
		return EX_UNAVAILABLE;
	}

	/* chop the key at the first newline found */
	for (p = key, c = 0; c < s.st_size; p++, c++)
	{
		if (*p == '\n')
		{
			*p = '\0';
			break;
		}
	}

	fclose(f);

	/* load allow list */
	if (allowlist != NULL)
	{
		RELIST new;
		char buf[BUFRSZ + 1];
		char restr[BUFRSZ + 1];

		f = fopen(allowlist, "r");
		if (f == NULL)
		{
			fprintf(stderr, "%s: %s: fopen(): %s\n", progname,
			        allowlist, strerror(errno));
			free(key);
			return EX_UNAVAILABLE;
		}

		memset(buf, '\0', sizeof buf);
		while (fgets(buf, sizeof buf - 1 ,f) != NULL)
		{
			for (p = buf; *p != '\0'; p++)
			{
				if (*p == '\n' || *p == '#')
				{
					*p = '\0';
					break;
				}
			}

			if (batv_isblank(buf))
				continue;

			batv_mkregexp(buf, restr, sizeof restr);

			new = (RELIST) malloc(sizeof *new);
			if (new == NULL)
			{
				fprintf(stderr, "%s: malloc(): %s\n",
				        progname, strerror(errno));
				fclose(f);
				return EX_UNAVAILABLE;
			}

			status = regcomp(&new->re_re, restr, REG_ICASE);
			if (status != 0)
			{
				char errbuf[BUFRSZ + 1];

				memset(errbuf, '\0', sizeof errbuf);
				(void) regerror(status, &new->re_re, errbuf,
				                sizeof errbuf);

				fprintf(stderr, "%s: regcomp(): %s\n",
				        progname, errbuf);
				fclose(f);
				return EX_UNAVAILABLE;
			}

			new->re_next = allow;
			allow = new;
		}

		if (ferror(f))
		{
			fprintf(stderr, "%s: %s: fgets(): %s\n", progname,
			        allowlist, strerror(errno));
			free(key);
			fclose(f);
			return EX_UNAVAILABLE;
		}

		fclose(f);
	}

	/* load sign list */
	if (ilist != NULL)
	{
		HOSTLIST new;
		struct in_addr addr;
		char *slash;
		char buf[BUFRSZ + 1];

		f = fopen(ilist, "r");
		if (f == NULL)
		{
			fprintf(stderr, "%s: %s: fopen(): %s\n", progname,
			        ilist, strerror(errno));
			free(key);
			return EX_UNAVAILABLE;
		}

		memset(buf, '\0', sizeof buf);
		while (fgets(buf, sizeof buf - 1 ,f) != NULL)
		{
			for (p = buf; *p != '\0'; p++)
			{
				if (*p == '\n' || *p == '#')
				{
					*p = '\0';
					break;
				}
			}

			if (batv_isblank(buf))
				continue;

			new = (HOSTLIST) malloc(sizeof *new);
			if (new == NULL)
			{
				fprintf(stderr, "%s: malloc(): %s\n",
				        progname, strerror(errno));
				fclose(f);
				return EX_UNAVAILABLE;
			}
			memset(new, '\0', sizeof *new);

			slash = strchr(buf, '/');
			if (slash != NULL)
				*slash = '\0';

			addr.s_addr = inet_addr(buf);
			if (addr.s_addr == INADDR_NONE)
			{
				new->host_name = strdup(buf);
			}
			else
			{
				if (slash == NULL)
				{
					new->host_addr.s_addr = addr.s_addr;
					new->host_mask.s_addr = INADDR_BROADCAST;
				}
				else
				{
					int bits;
					char *q;
					struct in_addr mask;

					mask.s_addr = 0;

					bits = strtoul(slash + 1, &q, 10);
					if (*q != '\0')
					{
						mask.s_addr = inet_addr(slash + 1);
						if (mask.s_addr == INADDR_NONE)
						{
							fprintf(stderr,
							        "%s: %s: invalid CIDR specification `%s/%s'\n",
							        progname,
							        ilist, buf,
							        slash + 1);
						}
						fclose(f);
						return EX_UNAVAILABLE;
					}
					else
					{
						int n;

						for (n = 31;
						     bits > 0;
						     bits--, n--)
							mask.s_addr |= htonl(1 << n);
					}

					new->host_addr.s_addr = addr.s_addr;
					new->host_mask.s_addr = mask.s_addr;
				}
			}

			new->host_next = signlist;
			signlist = new;
		}

		if (ferror(f))
		{
			fprintf(stderr, "%s: %s: fgets(): %s\n", progname,
			        ilist, strerror(errno));
			free(key);
			fclose(f);
			return EX_UNAVAILABLE;
		}

		fclose(f);
	}
	else
	{
		HOSTLIST new;

		new = (HOSTLIST) malloc(sizeof *new);
		if (new == NULL)
		{
			fprintf(stderr, "%s: malloc(): %s\n",
			        progname, strerror(errno));
			fclose(f);
			return EX_UNAVAILABLE;
		}

		new->host_name = strdup(LOCALHOST);
		new->host_addr.s_addr = htonl(INADDR_LOOPBACK);
		new->host_mask.s_addr = htonl(INADDR_BROADCAST);
		new->host_next = NULL;

		signlist = new;
	}

	/* store domain list */
	if (domainlist != NULL)
	{
		DOMLIST new;

		if (domainlist[0] == '/' && access(domainlist, F_OK) == 0)
		{
			char line[BUFRSZ + 1];

			f = fopen(domainlist, "r");
			if (f == NULL)
			{
				fprintf(stderr, "%s: %s: fopen(): %s\n",
				        progname, domainlist, strerror(errno));
				return EX_OSERR;
			}

			memset(line, '\0', sizeof line);
			while (fgets(line, BUFRSZ, f) != NULL)
			{
				for (p = line; *p != '\0'; p++)
				{
					if (*p == '\n' || *p == '#')
					{
						*p = '\0';
						break;
					}
				}

				batvf_trimspaces(line);
				if (strlen(line) == 0)
					continue;

				new = (DOMLIST) malloc(sizeof *new);
				if (new == NULL)
				{
					fprintf(stderr,
					        "%s: malloc(): %s\n",
					        progname, strerror(errno));
					fclose(f);
					return EX_OSERR;
				}

				new->dom_name = strdup(line);
				new->dom_next = domains;
				domains = new;
			}

			fclose(f);
		}
		else
		{
			char *tmp;

			tmp = strdup(domainlist);
			if (tmp == NULL)
			{
				fprintf(stderr, "%s: strdup(): %s\n",
				        progname, strerror(errno));
				fclose(f);
				return EX_OSERR;
			}

			for (p = strtok(tmp, ",");
			     p != NULL;
			     p = strtok(NULL, ","))
			{
				new = (DOMLIST) malloc(sizeof *new);
				if (new == NULL)
				{
					fprintf(stderr,
					        "%s: malloc(): %s\n",
					        progname, strerror(errno));
					fclose(f);
					return EX_OSERR;
				}

				new->dom_name = strdup(p);
				new->dom_next = domains;
				domains = new;
			}
		}
	}

	/* store MTA name list */
	if (mtanames != NULL)
	{
		int n = 1;
		char *tmp;

		tmp = strdup(mtanames);
		if (tmp == NULL)
		{
			fprintf(stderr, "%s: strdup(): %s\n",
			        progname, strerror(errno));
			return EX_UNAVAILABLE;
		}

		for (p = tmp; *p != '\0'; p++)
		{
			if (*p == ',')
				n++;
		}

		mtas = (char **) malloc((n + 1) * sizeof(char *));

		if (mtas == NULL)
		{
			fprintf(stderr, "%s: malloc(): %s\n",
			        progname, strerror(errno));
			return EX_UNAVAILABLE;
		}

		n = 0;
		for (p = strtok(tmp, ",");
		     p != NULL;
		     p = strtok(NULL, ","))
		{
			mtas[n] = p;
			n++;
		}

		mtas[n] = NULL;
	}

	/* store macro list */
	if (macrolist != NULL)
	{
		int n = 1;
		char *tmp;

		tmp = strdup(macrolist);
		if (tmp == NULL)
		{
			fprintf(stderr, "%s: strdup(): %s\n",
			        progname, strerror(errno));
			return EX_UNAVAILABLE;
		}

		for (p = tmp; *p != '\0'; p++)
		{
			if (*p == ',')
				n++;
		}

		macros = (char **) malloc((n + 1) * sizeof(char *));
		values = (char **) malloc((n + 1) * sizeof(char *));

		if (macros == NULL || values == NULL)
		{
			fprintf(stderr, "%s: malloc(): %s\n",
			        progname, strerror(errno));
			return EX_UNAVAILABLE;
		}

		n = 0;
		for (p = strtok(tmp, ",");
		     p != NULL;
		     p = strtok(NULL, ","))
		{
			macros[n] = p;
			values[n] = strchr(p, '=');
			if (values[n] != NULL)
			{
				*(values[n]) = '\0';
				values[n] += 1;
			}
			n++;
		}
		macros[n] = NULL;
		values[n] = NULL;
	}

#ifdef DEBUG
	return batv_debug();

#else /* DEBUG */

	/* Change user if appropriate */
	if (user != NULL)
	{
		struct passwd* pw;

		pw = getpwnam(user);
		if (pw == NULL)
		{
			uid_t uid;

			uid = atoi(user);
			if (uid != 0 && uid != LONG_MIN && uid != LONG_MAX)
				pw = getpwuid(uid);

			if (pw == NULL)
			{
				fprintf(stderr, "%s: no such user `%s'\n",
					progname, user);
				return EX_DATAERR;
			}
		}

		(void) endpwent();

		if (setuid(pw->pw_uid) != 0)
		{
			fprintf(stderr, "%s: setuid(): %s\n", progname,
				strerror(errno));
			return EX_NOPERM;
		}
	}
 
	/* Activate logging */
	if (dolog)
		batv_init_syslog(facility);

	if (autorestart)
	{
		bool quitloop = FALSE;
		int restarts = 0;
		pid_t pid;
		pid_t wpid;
		struct sigaction sa;

		if (dofork)
		{
			pid = fork();
			switch (pid)
			{
			  case -1:
				if (dolog)
				{
					int saveerrno;

					saveerrno = errno;

					syslog(LOG_ERR, "fork(): %s",
					       strerror(errno));

					errno = saveerrno;
				}

				fprintf(stderr, "%s: fork(): %s\n",
				        progname, strerror(errno));

				return EX_OSERR;

			  case 0:
				batv_stdio();
				break;

			  default:
				return EX_OK;
			}
		}

		if (pidfile != NULL)
		{
			f = fopen(pidfile, "w");
			if (f != NULL)
			{
				fprintf(f, "%ld\n", (long) getpid());
				(void) fclose(f);
			}
			else
			{
				if (dolog)
				{
					syslog(LOG_ERR,
					       "can't write pid to %s: %s",
					       pidfile, strerror(errno));
				}
			}
		}

		sa.sa_handler = batv_sighandler;
		sigemptyset(&sa.sa_mask);
		sigaddset(&sa.sa_mask, SIGHUP);
		sigaddset(&sa.sa_mask, SIGINT);
		sigaddset(&sa.sa_mask, SIGTERM);
		sa.sa_flags = 0;

		if (sigaction(SIGHUP, &sa, NULL) != 0 ||
		    sigaction(SIGINT, &sa, NULL) != 0 ||
		    sigaction(SIGTERM, &sa, NULL) != 0)
		{
			if (dolog)
			{
				syslog(LOG_ERR, "[parent] sigaction(): %s",
				       strerror(errno));
			}
		}

		while (!quitloop)
		{
			status = batvf_socket_cleanup(conn);
			if (status != 0)
			{
				if (dolog)
				{
					syslog(LOG_ERR,
					       "[parent] socket cleanup failed: %s",
					       strerror(status));
				}
				return EX_UNAVAILABLE;
			}

			pid = fork();
			switch (pid)
			{
			  case -1:
				if (dolog)
				{
					syslog(LOG_ERR, "fork(): %s",
					       strerror(errno));
				}

				return EX_OSERR;

			  case 0:
				sa.sa_handler = SIG_DFL;

				if (sigaction(SIGHUP, &sa, NULL) != 0 ||
				    sigaction(SIGINT, &sa, NULL) != 0 ||
				    sigaction(SIGTERM, &sa, NULL) != 0)
				{
					if (dolog)
					{
						syslog(LOG_ERR,
						       "[child] sigaction(): %s",
						       strerror(errno));
					}
				}

				quitloop = TRUE;
				break;

			  default:
				for (;;)
				{
					wpid = wait(&status);

					if (wpid == -1 && errno == EINTR)
					{
						if (die)
						{
							batv_killchild(pid,
							               diesig);

							while (wpid != pid)
								wpid = wait(&status);

							if (pidfile != NULL)
								(void) unlink(pidfile);

							exit(EX_OK);
						}
					}

					if (pid != wpid)
						continue;

					if (wpid != -1 && dolog)
					{
						if (WIFSIGNALED(status))
						{
							syslog(LOG_NOTICE,
							       "terminated with signal %d, restarting",
							       WTERMSIG(status));
						}
						else if (WIFEXITED(status))
						{
							syslog(LOG_NOTICE,
							       "exited with status %d, restarting",
							       WEXITSTATUS(status));
						}
					}

					break;
				}
				break;
			}

			if (maxrestarts > 0 && restarts >= maxrestarts)
			{
				if (dolog)
				{
					syslog(LOG_ERR,
					       "maximum restart count exceeded");
				}

				return EX_UNAVAILABLE;
			}

			if (maxrestartrate_n > 0 &&
			    maxrestartrate_t > 0 &&
			    !batvf_restart_check(0, maxrestartrate_t))
			{
				if (dolog)
				{
					syslog(LOG_ERR,
					       "maximum restart rate exceeded");
				}

				return EX_UNAVAILABLE;
			}

			restarts++;
		}
	}

	/* register with the milter interface */
	if (smfi_register(smfilter) == MI_FAILURE)
	{
		if (dolog)
			syslog(LOG_ERR, "smfi_register() failed");

		fprintf(stderr, "%s: smfi_register() failed\n", progname);

		return EX_UNAVAILABLE;
	}

	/* try to establish the socket */
	(void) smfi_setconn(conn);
	if (smfi_opensocket(TRUE) == MI_FAILURE)
	{
		if (dolog)
			syslog(LOG_ERR, "smfi_opensocket() failed");

		fprintf(stderr, "%s: smfi_opensocket() failed\n", progname);

		return EX_UNAVAILABLE;
	}

	if (!autorestart && dofork)
	{
		int save_errno;
		pid_t pid;

		pid = fork();
		switch (pid)
		{
		  case -1:
			save_errno = errno;
			(void) fprintf(stderr, "%s: fork(): %s\n",
				       progname, strerror(save_errno));
			if (dolog)
			{
				errno = save_errno;
				syslog(LOG_ERR, "fork(): %s", strerror(errno));
			}
			return EX_OSERR;

		  case 0:
			break;

		  default:
			return EX_OK;
		}
	}

	/*
	**  setsid() works on POSIX only.  Fortunately, that's
	**  all we need for now.  Check sendmail/conf.c for a
	**  portable version if needed.
	*/

	(void) setsid();

	if (!autorestart && !dofork)
	{
		/* redirect stdin, stdout, stderr to /dev/null */
		devnull = open(_PATH_DEVNULL, O_RDWR, 0);
		if (devnull < 0)
		{
			fprintf(stderr, "%s: %s: open(): %s\n", progname,
			        _PATH_DEVNULL, strerror(errno));
		}

		status = dup2(devnull, 0);
		if (status != -1)
			status = dup2(devnull, 1);
		if (status != -1)
			status = dup2(devnull, 2);
		if (status == -1)
		{
			fprintf(stderr, "%s: %s: dup2(): %s\n",
			        progname, _PATH_DEVNULL,
			        strerror(errno));
		}

		close(devnull);
	}

	/* write out the pid */
	if (!autorestart && pidfile != NULL)
	{
		f = fopen(pidfile, "w");
		if (f != NULL)
		{
			fprintf(f, "%u\n", getpid());
			(void) fclose(f);
		}
		else if (dolog)
		{
			syslog(LOG_ERR, "can't write pid to %s: %s",
			       pidfile, strerror(errno));
		}
	}

	/* call the milter mainline */
	if (dolog)
	{
		size_t n;
		char argstr[MAXARGV];
		char *end;

		memset(argstr, '\0', sizeof argstr);
		end = &argstr[sizeof argstr - 1];
		n = sizeof argstr;
		for (c = 1, p = argstr; c < argc && p < end; c++)
		{
			if (strchr(argv[c], ' ') != NULL)
			{
				status = snprintf(p, n, "%s \"%s\"",
				                  c == 1 ? "args:" : "",
				                  argv[c]);
			}
			else
			{
				status = snprintf(p, n, "%s %s",
				                  c == 1 ? "args:" : "",
				                  argv[c]);
			}

			p += status;
			n -= status;
		}

		syslog(LOG_INFO, "%s v%s starting (%s)", BATV_PRODUCT,
		       BATV_VERSION, argstr);
	}

	status = smfi_main();

	if (!autorestart && pidfile != NULL)
		(void) unlink(pidfile);

	return status;

#endif /* DEBUG */
}
